@@ -1,15 +1,20 @@
 """Use the GraphQL api to grab issues/PRs that match a query."""
 import datetime
-import dateutil
-import pytz
-import requests
+import os
+import sys
 import urllib
 from pathlib import Path
+from subprocess import PIPE
+from subprocess import run
 
-from .graphql import GitHubGraphQlQuery
-from .cache import _cache_data
-import pandas as pd
+import dateutil
 import numpy as np
+import pandas as pd
+import pytz
+import requests
+
+from .cache import _cache_data
+from .graphql import GitHubGraphQlQuery
 
 
 # The tags and description to use in creating subsets of PRs
@@ -21,7 +26,7 @@ TAGS_METADATA_BASE = {
     },
     "enhancement": {
         "tags": ["enhancement", "enhancements"],
-        "pre": ["NEW", "ENH", "ENHANCEMENT", "IMPROVE"],
+        "pre": ["ENH", "ENHANCEMENT", "IMPROVE", "IMP"],
         "description": "Enhancements made",
     },
     "bug": {
@@ -99,9 +104,20 @@ def get_activity(
         # We have just org
         search_query = f"user:{org}"
 
+    auth = auth or os.environ.get("GITHUB_ACCESS_TOKEN")
+    if not auth:
+        raise ValueError(
+            "Either the environment variable GITHUB_ACCESS_TOKEN or the "
+            "--auth flag or must be used to pass a Personal Access Token "
+            "needed by the GitHub API. You can generate a token at "
+            "https://github.com/settings/tokens/new. Note that while "
+            "working with a public repository, you don’t need to set any "
+            "scopes on the token you create."
+        )
+
     # Figure out dates for our query
-    since_dt, since_is_git_ref = _get_datetime_and_type(org, repo, since)
-    until_dt, until_is_git_ref = _get_datetime_and_type(org, repo, until)
+    since_dt, since_is_git_ref = _get_datetime_and_type(org, repo, since, auth)
+    until_dt, until_is_git_ref = _get_datetime_and_type(org, repo, until, auth)
     since_dt_str = f"{since_dt:%Y-%m-%dT%H:%M:%SZ}"
     until_dt_str = f"{until_dt:%Y-%m-%dT%H:%M:%SZ}"
 
@@ -112,7 +128,7 @@ def get_activity(
         search_query += f" type:{kind}"
 
     # Query for both opened and closed issues/PRs in this window
-    print(f"Running search query:\n{search_query}\n\n")
+    print(f"Running search query:\n{search_query}\n\n", file=sys.stderr)
     query_data = []
     for activity_type in ["created", "closed"]:
         ii_search_query = (
@@ -147,6 +163,8 @@ def generate_activity_md(
     include_issues=False,
     include_opened=False,
     strip_brackets=False,
+    heading_level=1,
+    branch=None,
 ):
     """Generate a markdown changelog of GitHub activity within a date window.
 
@@ -185,6 +203,12 @@ def generate_activity_md(
     strip_brackets : bool
         If True, strip any text between brackets at the beginning of the issue/PR title.
         E.g., [MRG], [DOC], etc.
+    heading_level : int
+        Base heading level to use.
+        By default, top-level heading is h1, sections are h2.
+        With heading_level=2 those are increased to h2 and h3, respectively.
+    branch : string | None
+        The branch or reference name to filter pull requests by
 
     Returns
     -------
@@ -195,8 +219,10 @@ def generate_activity_md(
     org, repo = _parse_target(target)
 
     # If no since parameter is given, find the name of the latest release
+    # using the _local_ git repostory
+    # TODO: Check that local repo matches org/repo
     if since is None:
-        since = _get_latest_tag(org, repo)
+        since = _get_latest_tag()
 
     # Grab the data according to our query
     data = get_activity(
@@ -247,6 +273,15 @@ def generate_activity_md(
     )
     data["kind"] = data["url"].map(lambda a: "issue" if "issues/" in a else "pr")
 
+    # Filter the PRs by branch (or ref) if given
+    if branch is not None:
+        index_names = data[
+            (data["kind"] == "pr") & (data["baseRefName"] != branch)
+        ].index
+        data.drop(index_names, inplace=True)
+        if data.empty:
+            return
+
     # Separate into closed and opened
     until_dt_str = data.until_dt_str
     since_dt_str = data.since_dt_str
@@ -289,14 +324,18 @@ def generate_activity_md(
     # Initialize our tags with empty metadata
     for key, vals in tags_metadata.items():
         vals.update(
-            {"mask": None, "md": [], "data": None,}
+            {
+                "mask": None,
+                "md": [],
+                "data": None,
+            }
         )
 
     # Separate out items by their tag types
     for kind, kindmeta in tags_metadata.items():
         # First find the PRs based on tag
         mask = closed_prs["labels"].map(
-            lambda a: any(ii in jj for ii in kindmeta["tags"] for jj in a)
+            lambda a: any(ii == jj for ii in kindmeta["tags"] for jj in a)
         )
         # Now find PRs based on prefix
         mask_pre = closed_prs["title"].map(
@@ -349,11 +388,13 @@ def generate_activity_md(
     # Generate the markdown
     prs = tags_metadata
 
+    extra_head = "#" * (heading_level - 1)
+
     for kind, items in prs.items():
         n_orgs = len(items["data"]["org"].unique())
         for org, idata in items["data"].groupby("org"):
             if n_orgs > 1:
-                items["md"].append(f"## {org}")
+                items["md"].append(f"{extra_head}## {org}")
                 items["md"].append("")
 
             for irow, irowdata in items["data"].iterrows():
@@ -361,11 +402,11 @@ def generate_activity_md(
                 ititle = irowdata["title"]
                 if strip_brackets and ititle.strip().startswith("[") and "]" in ititle:
                     ititle = ititle.split("]", 1)[-1].strip()
-                this_md = f"* {ititle} [#{irowdata['number']}]({irowdata['url']}) ([@{author}](https://github.com/{author}))"
+                this_md = f"- {ititle} [#{irowdata['number']}]({irowdata['url']}) ([@{author}](https://github.com/{author}))"
                 items["md"].append(this_md)
 
     # Get functional GitHub references: any git reference or master@{YY-mm-dd}
-    if not data.since_is_git_ref:
+    if closed_prs.size > 0 and not data.since_is_git_ref:
         since = f"master@{{{data.since_dt:%Y-%m-%d}}}"
         closest_date_start = closed_prs.loc[
             abs(
@@ -377,7 +418,7 @@ def generate_activity_md(
     else:
         since_ref = since
 
-    if not data.until_is_git_ref:
+    if closed_prs.size > 0 and not data.until_is_git_ref:
         until = f"master@{{{data.until_dt:%Y-%m-%d}}}"
         closest_date_stop = closed_prs.loc[
             abs(
@@ -392,11 +433,16 @@ def generate_activity_md(
     changelog_url = f"https://github.com/{org}/{repo}/compare/{since_ref}...{until_ref}"
 
     # Build the Markdown
-    md = [f"# {since}...{until}", f"([full changelog]({changelog_url}))", ""]
+    md = [
+        f"{extra_head}# {since}...{until}",
+        "",
+        f"([full changelog]({changelog_url}))",
+    ]
     for kind, info in prs.items():
         if len(info["md"]) > 0:
             md += [""]
-            md.append(f"## {info['description']}")
+            md.append(f"{extra_head}## {info['description']}")
+            md += [""]
             md += info["md"]
 
     # Add a list of author contributions
@@ -408,10 +454,12 @@ def generate_activity_md(
     contributor_md = " | ".join(all_contributor_links)
     gh_contributors_link = f"https://github.com/{org}/{repo}/graphs/contributors?from={data.since_dt:%Y-%m-%d}&to={data.until_dt:%Y-%m-%d}&type=c"
     md += [""]
-    md += ["## Contributors to this release"]
+    md += [f"{extra_head}## Contributors to this release"]
+    md += [""]
     md += [f"([GitHub contributors page for this release]({gh_contributors_link}))"]
     md += [""]
     md += [contributor_md]
+    md += [""]
     md = "\n".join(md)
     return md
 
@@ -459,8 +507,21 @@ def extract_comments(comments):
 
 
 def _parse_target(target):
+    """
+    Returns (org, repo) based on input such as:
+
+    - executablebooks
+    - executablebooks/jupyter-book
+    - http(s)://github.com/executablebooks
+    - http(s)://github.com/executablebooks/jupyter-book(.git)
+    - git@github.com:executablebooks/jupyter-book(.git)
+    """
     if target.startswith("http"):
         target = target.split("github.com/")[-1]
+    elif "@github.com:" in target:
+        target = target.split("@github.com:")[-1]
+    if target.endswith(".git"):
+        target = target.rsplit(".git", 1)[0]
     parts = target.split("/")
     if len(parts) == 2:
         org, repo = parts
@@ -474,37 +535,42 @@ def _parse_target(target):
     return org, repo
 
 
-def _get_datetime_and_type(org, repo, datetime_or_git_ref):
-    """Return a datetime object and bool."""
+def _get_datetime_and_type(org, repo, datetime_or_git_ref, auth):
+    """Return a datetime object and bool indicating if it is a git reference or
+    not."""
 
-    is_git_ref = False
+    # Default a blank datetime_or_git_ref to current UTC time, which makes sense
+    # to set the until flags default value.
     if datetime_or_git_ref is None:
         dt = datetime.datetime.now().astimezone(pytz.utc)
-    else:
+        return (dt, False)
+
+    try:
+        dt = _get_datetime_from_git_ref(org, repo, datetime_or_git_ref, auth)
+        return (dt, True)
+    except Exception as ref_error:
         try:
             dt = dateutil.parser.parse(datetime_or_git_ref)
-        except ValueError as e:
-            if repo:
-                dt = _get_datetime_from_git_ref(org, repo, datetime_or_git_ref)
-                is_git_ref = True
-            else:
-                raise e
-
-    return dt, is_git_ref
+            return (dt, False)
+        except Exception as datetime_error:
+            raise ValueError(
+                "{0} not found as a ref or valid date format".format(
+                    datetime_or_git_ref
+                )
+            )
 
 
-def _get_datetime_from_git_ref(org, repo, ref):
+def _get_datetime_from_git_ref(org, repo, ref, auth):
     """Return a datetime from a git reference."""
-
-    response = requests.get(f"https://api.github.com/repos/{org}/{repo}/commits/{ref}")
+    headers = {"Authorization": "Bearer %s" % auth}
+    url = f"https://api.github.com/repos/{org}/{repo}/commits/{ref}"
+    response = requests.get(url, headers=headers)
     response.raise_for_status()
     return dateutil.parser.parse(response.json()["commit"]["committer"]["date"])
 
 
-def _get_latest_tag(org, repo):
-    """Return the latest tag name for a given repository."""
-    response = requests.get(f"https://api.github.com/repos/{org}/{repo}/git/refs/tags")
-    response.raise_for_status()
-    tags = response.json()
-    latest_tag = list(tags)[-1]
-    return latest_tag["ref"].split("/tags/")[-1]
+def _get_latest_tag():
+    """Return the latest tag name for a given repository by querying the local repo."""
+    out = run("git describe --tags".split(), stdout=PIPE)
+    tag = out.stdout.decode().rsplit("-", 2)[0]
+    return tag
